from reling.app.app import app
from reling.gpt_log import get_log

__all__ = [
    'gpt_log',
]

ITEM_DIVIDER = '-' * 10


@app.command(hidden=True)
def gpt_log() -> None:
    """Display the GPT log from the last application run."""
    for item_index, item in enumerate(get_log()):
        if item_index > 0:
            print(ITEM_DIVIDER + '\n')
        print(f'Prompt:\n"""\n{item.prompt}\n"""\n')
        print(f'Temperature:\n{item.temperature}\n')
        print(f'Response:\n"""\n{item.response}\n"""\n')



from itertools import zip_longest
from pathlib import Path
from typing import Callable, Generator, Iterable

from reling.asr import ASRClient
from reling.config import MAX_SCORE
from reling.db.enums import ContentCategory
from reling.db.models import Language
from reling.helpers.colors import default, fade, red
from reling.helpers.input import get_input, ScannerParams, TranscriberParams
from reling.helpers.output import output, SentenceData
from reling.scanner import Scanner
from reling.tts import TTSVoiceClient
from reling.types import DialogueExchangeData
from reling.utils.transformers import get_numbering_prefix
from .types import ExchangeWithTranslation, ScoreWithSuggestion, SentenceWithTranslation

__all__ = [
    'collect_translations',
]

HIDDEN_TEXT = '(...)'
TRANSLATION = 'Translation'
ATTEMPT = 'Attempt #{n}'
ATTEMPT_TITLE_DELIMITER = '-' * 3
PROMPT_SUFFIX = ': '
SCORE_SUFFIX = f' ({{score}}/{MAX_SCORE})'

SKIPPED = fade('(skipped)')


def collect_translations(
        category: ContentCategory,
        items: Iterable[str | DialogueExchangeData],
        original_translations: Iterable[str | DialogueExchangeData],
        skipped_indices: set[int],
        target_language: Language,
        source_tts: TTSVoiceClient | None,
        target_speaker_tts: TTSVoiceClient | None,
        asr: ASRClient | None,
        scanner: Scanner | None,
        hide_prompts: bool,
        previous_attempts: list[list[SentenceWithTranslation | ExchangeWithTranslation]],
        previous_scores: list[list[ScoreWithSuggestion]],
        storage: Path,
        on_pause: Callable[[], None],
        on_resume: Callable[[], None],
) -> Generator[SentenceWithTranslation | ExchangeWithTranslation, None, None]:
    """Collect the translations of text sentences or user turns in a dialogue."""
    attempt_title = ATTEMPT.format(n=len(previous_attempts) + 1) if previous_attempts else None
    if attempt_title:
        print(attempt_title)
        print(ATTEMPT_TITLE_DELIMITER)
    is_dialogue = category == ContentCategory.DIALOGUE
    speaker_translations: list[str] = []
    collected: list[str] = []
    for index, (item, original_translation) in enumerate(zip(items, original_translations)):
        print_prefix = get_numbering_prefix(index)
        if is_dialogue:
            speaker_translations.append(original_translation.speaker)
        if index in skipped_indices:
            if is_dialogue:
                output(SentenceData(
                    print_text=original_translation.speaker,
                ))
            output(SentenceData(
                print_text=item.user if is_dialogue else item,
                print_prefix=print_prefix,
            ))
            output(SentenceData(SKIPPED))
            translation = None
            collected.append(original_translation.user if is_dialogue else original_translation)
        else:
            if is_dialogue:
                output(SentenceData.from_tts(
                    original_translation.speaker,
                    target_speaker_tts,
                ))
            output(SentenceData.from_tts(
                item.user if is_dialogue else item,
                source_tts,
                print_text=HIDDEN_TEXT if hide_prompts else None,
                print_prefix=print_prefix,
            ))
            if previous_attempts:
                for attempt_index, (attempts, scores) in enumerate(zip(previous_attempts, previous_scores)):
                    output(SentenceData(
                        print_text=(default(attempts[index].input.text)
                                    + red(SCORE_SUFFIX.format(score=scores[index].score))),
                        print_prefix=ATTEMPT.format(n=attempt_index + 1) + PROMPT_SUFFIX,
                    ))
            translation = get_input(
                on_pause=on_pause,
                on_resume=on_resume,
                prompt=(attempt_title or TRANSLATION) + PROMPT_SUFFIX,
                input_suggestions=[attempts[index].input.text for attempts in previous_attempts],
                transcriber_params=TranscriberParams(
                    transcribe=asr.get_transcriber(target_language, '\n'.join(
                        (turn for exchange in zip_longest(speaker_translations, collected) for turn in exchange if turn)
                        if is_dialogue
                        else collected
                    )),
                    storage=storage,
                ) if asr else None,
                scanner_params=ScannerParams(
                    scanner=scanner,
                    language=target_language,
                ) if scanner else None,
            )
            collected.append(translation.text)
        yield (ExchangeWithTranslation(item, translation)
               if is_dialogue
               else SentenceWithTranslation(item, translation))
        print()



from .typer import typer_raise_import

__all__ = [
    'ensure_audio',
]


def ensure_audio() -> None:
    """Ensure the tool is installed with the necessary audio dependencies."""
    try:
        import pyaudio
    except ImportError:
        typer_raise_import('Audio dependencies')



import re
from typing import Generator
from unicodedata import category, normalize

__all__ = [
    'capitalize_first_char',
    'char_range',
    'is_cj',
    'is_punctuation',
    'is_whitespace',
    'replace_prefix_casing',
    'tokenize',
    'universal_normalize',
]

CJ = re.compile(r'[\u4E00-\u9FFF\u3040-\u30FF]')


def universal_normalize(string: str) -> str:
    """Normalize a string using Unicode NFC and replace specific characters."""
    return normalize('NFC', string).replace('’', "'")


def capitalize_first_char(string: str) -> str:
    """Capitalize the first character of a string."""
    return string[:1].upper() + string[1:]


def replace_prefix_casing(string: str, prefix: str) -> str:
    """Replace the prefix of a string with the casing of the given prefix."""
    return prefix + string[len(prefix):] if string.lower().startswith(prefix.lower()) else string


def is_punctuation(char: str) -> bool:
    """Return whether the character is a punctuation symbol."""
    return category(char).startswith('P')


def is_whitespace(char: str) -> bool:
    """Return whether the character is a whitespace character."""
    return category(char).startswith('Z')


def is_cj(char: str) -> bool:
    """Return whether the character is a Chinese or Japanese character."""
    return CJ.fullmatch(char) is not None


def tokenize(string: str, *, punctuation: bool = True, whitespace: bool = True, cj: bool = True) -> list[str]:
    """Tokenize a string into words, punctuation, whitespace, and individual CJ characters."""
    tokens: list[str] = []
    current: list[str] = []
    for char in string:
        for (checker, should_include) in [
            (is_punctuation, punctuation),
            (is_whitespace, whitespace),
            (is_cj, cj),
        ]:
            if checker(char):
                if current:
                    tokens.append(''.join(current))
                    current.clear()
                if should_include:
                    tokens.append(char)
                break
        else:
            current.append(char)
    if current:
        tokens.append(''.join(current))
    return tokens


def char_range(from_char: str, to_char: str) -> Generator[str, None, None]:
    """Generate characters from the given range, inclusive."""
    direction = 1 if from_char <= to_char else -1
    for char_code in range(ord(from_char), ord(to_char) + direction, direction):
        yield chr(char_code)



from __future__ import annotations

from lcs2 import lcs_length

FUZZY_RATIO = 0.61


class FuzzyWord:
    """
    A class representing a word for fuzzy comparison.
    Two words are considered equal if they share at least FUZZY_RATIO of their characters, ignoring case.
    """
    _normalized: str

    def __init__(self, word: str) -> None:
        self._normalized = word.lower()

    def __eq__(self, other: FuzzyWord) -> bool:
        return (2 * lcs_length(self._normalized, other._normalized)
                >= (len(self._normalized) + len(other._normalized)) * FUZZY_RATIO)



from pathlib import Path
import shelve
from typing import Any

__all__ = [
    'delete_value',
    'get_value',
    'init_shelf',
    'set_value',
]

FILENAME: Path | None = None


def init_shelf(filename: Path) -> None:
    global FILENAME
    FILENAME = filename


def get_filename() -> str:
    if FILENAME is None:
        raise RuntimeError('Shelf is not initialized.')
    return str(FILENAME.absolute())


def get_value[T](key: str, default: T | None = None) -> T | None:
    with shelve.open(get_filename()) as db:
        return db.get(key, default)


def set_value(key: str, value: Any) -> None:
    with shelve.open(get_filename()) as db:
        db[key] = value


def delete_value(key: str) -> None:
    with shelve.open(get_filename()) as db:
        if key in db:
            del db[key]



from reling.app.app import app
from reling.app.default_content import set_default_content
from reling.app.types import CONTENT_ARG
from reling.db import single_session
from reling.helpers.typer import typer_raise
from reling.utils.time import now

__all__ = [
    'archive',
]


@app.command()
def archive(content: CONTENT_ARG) -> None:
    """Archive a text or dialogue."""
    set_default_content(content)

    if content.archived_at is not None:
        typer_raise('The content is already archived.')

    with single_session() as session:
        content.archived_at = now()
        session.commit()
    print(f'Archived "{content.id}".')



from reling.app.app import app
from reling.app.default_content import set_default_content
from reling.app.types import CONTENT_ARG
from reling.db import single_session
from reling.helpers.typer import typer_raise

__all__ = [
    'unarchive',
]


@app.command()
def unarchive(content: CONTENT_ARG) -> None:
    """Unarchive a text or dialogue."""
    set_default_content(content)

    if content.archived_at is None:
        typer_raise('The content is not archived.')

    with single_session() as session:
        content.archived_at = None
        session.commit()
    print(f'Unarchived "{content.id}".')



from __future__ import annotations
import base64
from contextlib import contextmanager
from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import cv2

from reling.db.models import Language
from reling.gpt import GPTClient
from reling.helpers.typer import typer_raise, typer_raise_import
from reling.types import Image
from reling.utils.transformers import strip

__all__ = [
    'Scanner',
    'ScannerManager',
    'ScannerParams',
]

IMAGE_EXTENSION = '.jpg'
IMAGE_URL_FORMAT = 'data:image/jpeg;base64,{image}'


@dataclass
class ScannerParams:
    camera_index: int
    gpt: GPTClient


class Scanner:
    _camera: cv2.VideoCapture
    _gpt: GPTClient

    def __init__(self, params: ScannerParams) -> None:
        try:
            import cv2
        except ImportError:
            typer_raise_import('OpenCV')
        self._camera = cv2.VideoCapture(params.camera_index)
        if not self._camera.isOpened():
            typer_raise(f'Camera with index {params.camera_index} is not available.')
        self._gpt = params.gpt

    def release(self) -> None:
        self._camera.release()

    def capture(self) -> Image:
        """Capture an image from the camera."""
        success, frame = self._camera.read()
        if not success:
            typer_raise('Failed to capture an image.')
        import cv2
        success, buffer = cv2.imencode(IMAGE_EXTENSION, frame)
        if not success:
            typer_raise('Failed to encode the image.')
        return Image(IMAGE_URL_FORMAT.format(image=base64.b64encode(buffer).decode('utf-8')))

    def process(self, image: Image, language: Language) -> str:
        """Process the image and extract text in the specified language."""
        return list(self._gpt.ask(
            '\n'.join([
                f'What is written in the following image? The text is in {language.name}.',
                f'It might be written across several lines, but it forms a single sentence or paragraph.',
                f'Please provide your answer in a single line. Say nothing else.',
                f'If no text in {language.name} is detected, write three dots.',
            ]),
            image=image,
            creative=False,
            transformers=[strip],
        ))[0]


class ScannerManager:
    _params: ScannerParams | None

    def __init__(self, params: ScannerParams | None) -> None:
        self._params = params

    @contextmanager
    def get_scanner(self) -> Scanner | None:
        if self._params is None:
            yield None
        else:
            scanner = Scanner(self._params)
            try:
                yield scanner
            finally:
                scanner.release()



from sqlalchemy import ForeignKey, Index
from sqlalchemy.orm import Mapped, mapped_column

from reling.db.base import Base
from .languages import Language

__all__ = [
    'GrammarCacheSentence',
    'GrammarCacheWord',
]


class GrammarCacheSentence(Base):
    __tablename__ = 'grammar_cache_sentences'

    id: Mapped[str] = mapped_column(primary_key=True)
    language_id: Mapped[str] = mapped_column(ForeignKey(Language.id))
    sentence: Mapped[str]

    __table_args__ = (
        Index('grammar_cache_sentence_language_sentence', 'language_id', 'sentence', unique=True),
    )


class GrammarCacheWord(Base):
    __tablename__ = 'grammar_cache_words'

    sentence_id: Mapped[str] = mapped_column(
        ForeignKey(GrammarCacheSentence.id, onupdate='CASCADE', ondelete='CASCADE'),
        primary_key=True,
    )
    index: Mapped[int] = mapped_column(primary_key=True)
    text: Mapped[str]
    lemma: Mapped[str]
    upos: Mapped[str]



from .consolidation import get_dialogue_exchanges, get_text_sentences
from .exceptions import TranslationExistsException
from .operation import translate_dialogue, translate_text

__all__ = [
    'get_dialogue_exchanges',
    'get_text_sentences',
    'translate_dialogue',
    'translate_text',
    'TranslationExistsException',
]



from dataclasses import dataclass
from datetime import datetime
from typing import cast

from reling.config import MAX_SCORE
from reling.db.models import Dialogue, DialogueExam, DialogueExamResult, Language, Text, TextExam, TextExamResult

__all__ = [
    'compute_pair_streaks',
    'compute_streaks',
    'LanguagePair',
    'Streak',
    'StreakTimeline',
]


@dataclass(frozen=True)
class LanguagePair:
    source_language: Language
    target_language: Language


@dataclass
class StreakTimeline:
    first: datetime
    last: datetime


@dataclass
class Streak:
    timeline: StreakTimeline | None = None
    count: int = 0
    ever_failed: bool = False


def update(streak: Streak, result: TextExamResult | DialogueExamResult, exam: TextExam | DialogueExam) -> None:
    """Update the streaks list with the given result."""
    if streak.ever_failed:
        return
    is_perfect = result.score == MAX_SCORE
    if not is_perfect:
        streak.ever_failed = True
        return
    streak.count += 1
    time = cast(datetime, exam.finished_at)
    if streak.timeline:
        streak.timeline.first = time
    else:
        streak.timeline = StreakTimeline(time, time)


def get_default_streaks(content: Text | Dialogue) -> list[Streak]:
    """Get the default streaks list."""
    return [Streak() for _ in range(content.size)]


def compute_streaks(
        content: Text | Dialogue,
        source_language: Language | None,
        target_language: Language | None,
) -> dict[LanguagePair, list[Streak]]:
    """Compute the sentence streaks for the given content and language pair."""
    streaks: dict[LanguagePair, list[Streak]] = {}
    for exam in cast(list[TextExam] | list[DialogueExam], content.exams):
        pair = LanguagePair(exam.source_language, exam.target_language)
        if pair != LanguagePair(source_language or exam.source_language,
                                target_language or exam.target_language):
            continue
        if pair not in streaks:
            streaks[pair] = get_default_streaks(content)
        for result in exam.results:
            update(streaks[pair][result.index], result, exam)
    return streaks


def compute_pair_streaks(
        content: Text | Dialogue,
        source_language: Language,
        target_language: Language,
) -> list[Streak]:
    """Compute the sentence streaks for the given content and language pair."""
    streaks = compute_streaks(content, source_language, target_language)
    pair = LanguagePair(source_language, target_language)
    return streaks[pair] if pair in streaks else get_default_streaks(content)



from abc import ABC, abstractmethod
from queue import Queue

__all__ = [
    'CharFeeder',
    'Feeder',
    'LineFeeder',
]


class Feeder(ABC):
    """
    Feeder can be fed chunks of text data. It then processes the data and provides it in a different form
    (say, line-by-line).
    """

    @abstractmethod
    def put(self, chunk: str) -> None:
        """Feed a chunk of data."""
        pass

    @abstractmethod
    def end(self) -> None:
        """Signal the end of the data."""
        pass

    @abstractmethod
    def get(self) -> str | None:
        """Get the next piece of data or None if there is no more data ready."""
        pass


class CharFeeder(Feeder):
    """Feeder that provides data character-by-character when the characters are ready."""

    _chars: Queue[str]

    def __init__(self) -> None:
        self._chars = Queue()

    def put(self, chunk: str) -> None:
        for char in chunk:
            self._chars.put(char)

    def end(self) -> None:
        pass

    def get(self) -> str | None:
        return self._chars.get() if not self._chars.empty() else None


class LineFeeder(Feeder):
    """
    Feeder that provides data line-by-line when the lines are ready.
    If the data ends with \n, the last empty line is not included.
    """

    _lines: Queue[str]
    _buffer: list[str]

    def __init__(self) -> None:
        self._lines = Queue()
        self._buffer = []

    def put(self, chunk: str) -> None:
        lines = chunk.split('\n')
        self._buffer.append(lines[0])
        if len(lines) > 1:
            self._lines.put(''.join(self._buffer))
            for line in lines[1:-1]:
                self._lines.put(line)
            self._buffer.clear()
            self._buffer.append(lines[-1])

    def end(self) -> None:
        if line := ''.join(self._buffer):
            self._lines.put(line)
        self._buffer.clear()

    def get(self) -> str | None:
        return self._lines.get() if not self._lines.empty() else None



from itertools import zip_longest
from typing import cast, Generator, Iterable, Iterator, overload

__all__ = [
    'extract_items',
    'intersperse',
    'group_items',
    'pair_items',
    'strict_zip',
]


def group_items[T](iterable: Iterable[T], group_size: int) -> Generator[tuple[T, ...], None, None]:
    """
    Group items from the iterable into tuples of the specified size.
    The last tuple is discarded if it is shorter than the specified size.
    """
    items = iter(iterable)
    for item in zip(*(items for _ in range(group_size))):
        yield item


def pair_items[T](iterable: Iterable[T]) -> Generator[tuple[T, T], None, None]:
    """Pair items from the iterable into tuples. The last item is discarded if there is an odd number of items."""
    return cast(Generator[tuple[T, T], None, None], group_items(iterable, 2))


def extract_items[T](iterable: Iterable[T], indices: Iterable[int]) -> Generator[T, None, None]:
    """Extract items from the iterable by the specified indices."""
    index_set = set(indices)
    for index, item in enumerate(iterable):
        if index in index_set:
            yield item


def intersperse[T](outer: Iterable[T], inner_with_indices: Iterable[tuple[T, int]]) -> Generator[T, None, None]:
    """
    Intersperse the inner iterable into the outer iterable at the specified indices;
    the inner iterable must be sorted by the indices.
    """
    nothing = object()

    outer_iter = iter(outer)
    inner_iter = iter(inner_with_indices)

    outer_value = next(outer_iter, nothing)
    inner_value, inner_index = next(inner_iter, (nothing, nothing))
    index = 0
    while True:
        if index == inner_index:
            yield inner_value
            inner_value, inner_index = next(inner_iter, (nothing, nothing))
        else:
            if outer_value is nothing:
                if inner_value is nothing:
                    return
                else:
                    raise ValueError('The outer iterable ended prematurely.')
            yield outer_value
            outer_value = next(outer_iter, nothing)
        index += 1


@overload
def strict_zip[T1, T2](
        exception: Exception,
        __iter1: Iterable[T1],
        __iter2: Iterable[T2],
) -> Iterator[tuple[T1, T2]]: ...


@overload
def strict_zip[T1, T2, T3](
        exception: Exception,
        __iter1: Iterable[T1],
        __iter2: Iterable[T2],
        __iter3: Iterable[T3],
) -> Iterator[tuple[T1, T2, T3]]: ...


@overload
def strict_zip[T1, T2, T3, T4](
        exception: Exception,
        __iter1: Iterable[T1],
        __iter2: Iterable[T2],
        __iter3: Iterable[T3],
        __iter4: Iterable[T4],
) -> Iterator[tuple[T1, T2, T3, T4]]: ...


def strict_zip(exception: Exception, *iterables: Iterable) -> Generator[tuple, None, None]:
    """Zip the iterables together, raising an exception if they are not the same length."""
    sentinel = object()
    for i, values in enumerate(zip_longest(*iterables, fillvalue=sentinel)):
        if sentinel in values:
            raise exception
        yield values



from dataclasses import dataclass
from pathlib import Path

__all__ = [
    'Input',
]


@dataclass
class Input:
    text: str
    audio: Path | None = None

    def __init__(self, text: str, audio: Path | None = None) -> None:
        self.text = text.strip()
        self.audio = audio



from datetime import datetime, timedelta

from reling.db import single_session
from reling.db.models import Dialogue, DialogueExam, DialogueExamResult, Language, Text, TextExam, TextExamResult
from reling.utils.ids import generate_id
from .types import ExchangeWithTranslation, ScoreWithSuggestion, SentenceWithTranslation

__all__ = [
    'save_exam',
]


def save_exam(
        content: Text | Dialogue,
        source_language: Language,
        target_language: Language,
        read_source: bool,
        read_target: bool,
        listened: bool,
        scanned: bool,
        started_at: datetime,
        finished_at: datetime,
        total_pause_time: timedelta,
        items: list[SentenceWithTranslation | ExchangeWithTranslation],
        results: list[ScoreWithSuggestion | None],
) -> TextExam | DialogueExam:
    """Save the results of a text or dialogue exam."""
    is_text = isinstance(content, Text)
    with single_session() as session:
        exam = (TextExam if is_text else DialogueExam)(
            id=generate_id(),
            content_id=content.id,
            source_language_id=source_language.id,
            target_language_id=target_language.id,
            read_source=read_source,
            read_target=read_target,
            listened=listened,
            scanned=scanned,
            started_at=started_at,
            finished_at=finished_at,
            total_pause_time=total_pause_time,
        )
        session.add(exam)
        for index, (item, result) in enumerate(zip(items, results)):
            if result is not None:
                session.add((TextExamResult if is_text else DialogueExamResult)(
                    exam_id=exam.id,
                    index=index,
                    answer=item.input.text,
                    suggested_answer=result.suggestion,
                    score=result.score,
                ))
        session.commit()
        return exam



from reling.app.app import app
from reling.app.db import get_db_path

__all__ = [
    'db',
]


@app.command()
def db() -> None:
    """Display the file path of the database."""
    print(get_db_path().absolute())



from enum import StrEnum
from typing import Any, Callable

from lcs2 import lcs_indices
from rich.text import Text

from reling.utils.strings import tokenize
from .colors import default, green, red
from .fuzzy_word import FuzzyWord

__all__ = [
    'DiffType',
    'highlight_diff',
]


class DiffType(StrEnum):
    CHAR = 'char'
    TOKEN = 'token'

    def get_tokenizer(self) -> Callable[[str], list[str]]:
        match self:
            case DiffType.CHAR:
                return lambda section: list(section)
            case DiffType.TOKEN:
                return lambda section: tokenize(section)
            case _:
                raise NotImplementedError

    def get_normalizer(self) -> Callable[[str], Any]:
        match self:
            case DiffType.CHAR:
                return lambda section: section
            case DiffType.TOKEN:
                return lambda section: FuzzyWord(section)
            case _:
                raise NotImplementedError


def highlight_diff(worse: str, better: str, diff_type: DiffType = DiffType.TOKEN) -> tuple[Text, Text]:
    """Return the formatted pair of strings, highlighting the difference between the two."""
    tokenizer, normalizer = diff_type.get_tokenizer(), diff_type.get_normalizer()
    worse_tokens, better_tokens = tokenizer(worse), tokenizer(better)
    lcs = lcs_indices(*([normalizer(token) for token in tokens] for tokens in (worse_tokens, better_tokens)))
    worse_segments: list[Text] = []
    better_segments: list[Text] = []

    worse_cursor = better_cursor = 0
    for worse_index, better_index in lcs + [(len(worse_tokens), len(better_tokens))]:
        worse_segments.append(red(''.join(worse_tokens[worse_cursor:worse_index])))
        better_segments.append(green(''.join(better_tokens[better_cursor:better_index])))
        if (worse_index, better_index) == (len(worse_tokens), len(better_tokens)):
            break
        match diff_type:
            case DiffType.CHAR:
                worse_segments.append(default(worse_tokens[worse_index]))
                better_segments.append(default(better_tokens[better_index]))
            case DiffType.TOKEN:
                worse_segment, better_segment = highlight_diff(
                    worse_tokens[worse_index],
                    better_tokens[better_index],
                    DiffType.CHAR,
                )
                worse_segments.append(worse_segment)
                better_segments.append(better_segment)
        worse_cursor = worse_index + 1
        better_cursor = better_index + 1

    return sum(worse_segments, Text('')), sum(better_segments, Text(''))



from dataclasses import dataclass

from reling.shelf import delete_value, get_value, set_value

__all__ = [
    'get_log',
    'GptLogItem',
    'log',
]

ITEMS_VAR_NAME = 'GPT_LOG_ITEM_{index}'
COUNT_VAR_NAME = 'GPT_LOG_COUNT'

CURRENT_RUN_COUNT = 0


@dataclass
class GptLogItem:
    prompt: str
    temperature: float
    response: str


def clear() -> None:
    """Remove all items from the GPT log."""
    for index in range(get_value(COUNT_VAR_NAME, 0)):
        delete_value(ITEMS_VAR_NAME.format(index=index))


def log(item: GptLogItem) -> None:
    """Append an item to the GPT log; if this is the first item, clear the log from the previous run first."""
    global CURRENT_RUN_COUNT
    if CURRENT_RUN_COUNT == 0:
        clear()
    set_value(ITEMS_VAR_NAME.format(index=CURRENT_RUN_COUNT), item)
    CURRENT_RUN_COUNT += 1
    set_value(COUNT_VAR_NAME, CURRENT_RUN_COUNT)


def get_log() -> list[GptLogItem]:
    """Get the GPT log items."""
    count = get_value(COUNT_VAR_NAME, 0)
    return [
        get_value(ITEMS_VAR_NAME.format(index=index))
        for index in range(count)
    ]



from reling.db.models import IdIndex

from reling.db import single_session

__all__ = [
    'find_ids_by_prefix',
]


def find_ids_by_prefix(prefix: str) -> list[str]:
    """Find content IDs that start with the given prefix, ordered by ID."""
    with single_session() as session:
        return [
            result.id
            for result in session.query(IdIndex)
            .where(IdIndex.id.startswith(prefix))
            .order_by(IdIndex.id)
            .all()
        ]
